package org.moeaframework.algorithm;

import java.io.Serializable;
import java.util.Comparator;

import org.moeaframework.core.FastNondominatedSorting;
import org.moeaframework.core.FitnessEvaluator;
import org.moeaframework.core.Initialization;

import org.moeaframework.core.Population;
import org.moeaframework.core.Problem;
import org.moeaframework.core.Selection;
import org.moeaframework.core.Solution;
import org.moeaframework.core.Variation;
import org.moeaframework.core.comparator.ChainedComparator;
import org.moeaframework.core.comparator.CrowdingComparator;
import org.moeaframework.core.comparator.FitnessComparator;
import org.moeaframework.core.comparator.NondominatedSortingComparator;
import org.moeaframework.core.comparator.ParetoDominanceComparator;
import org.moeaframework.core.comparator.RankComparator;
import org.moeaframework.core.operator.TournamentSelection;
import static org.moeaframework.core.NondominatedSorting.RANK_ATTRIBUTE;


public class DI1 extends AbstractEvolutionaryAlgorithm {

	

	private FitnessEvaluator fitnessEvaluator;
	
	/**
	 * The selection operator.
	 */
	private Selection selection;
	
	/**
	 * The variation operator.
	 */
	private Variation variation;

	private boolean initial;
	
	/**
	 * Constructs a new DI-1 instance.
	 * 
	 * @param problem the problem
	 * @param initialization the initialization operator
	 * @param variation the variation operator
	 * @param fitnessEvaluator the fitness evaluator
	 */
	public DI1(Problem problem, Initialization initialization,
			Variation variation, FitnessEvaluator fitnessEvaluator) {
		super(problem,
				new Population(),
				null,
				initialization);
		this.variation = variation;
		this.fitnessEvaluator = fitnessEvaluator;
		this.initial = false;
		
		if (fitnessEvaluator ==  null) {
			selection = new TournamentSelection(
					new NondominatedSortingComparator());
		} else {
			selection = new TournamentSelection(
					new NondominatedFitnessComparator());
		}

	}

	@Override
	protected void initialize() {
		super.initialize();
		
		// rank the solutions
		//rank and crowding distance(for removing duplicated solutions)
		new FastNondominatedSorting().evaluate(population);
		initial = false;
		
		//calculate  contribution
//		if (fitnessEvaluator != null) {
//			fitnessEvaluator.evaluate(population);
//		}
	}

	@Override
	protected void iterate() {

		int populationSize = population.size();
		Solution solution = population.get(populationSize-1);
		int lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
				
		if ((lastrank != 0) || (initial == false))
		{
			selection = new TournamentSelection(2, new ChainedComparator(
					new ParetoDominanceComparator(),
					new CrowdingComparator()));
			
			initial = true;
			Population offspring = new Population();
			
			while (offspring.size() < populationSize) {
				Solution[] parents = selection.select(variation.getArity(),
						population);

				offspring.addAll(variation.evolve(parents));
			}
			
			evaluateAll(offspring);

			population.addAll(offspring);
			
			new FastNondominatedSorting().evaluate(population);
			population.truncate(populationSize, 
					new NondominatedFitnessComparator1());
			
			
			solution = population.get(populationSize-1);
			lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
						
			if ((lastrank == 0))
			{
				fitnessEvaluator.evaluate(population);
		
			}
			
			//test code
			solution = population.get(populationSize-1);
			lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
			int[] ranknumtest = new int[lastrank+1];
			for (int i = 0; i<populationSize; i++)
			{
				int temp = (int) population.get(i).getAttribute(RANK_ATTRIBUTE);
				ranknumtest[temp]++; 
			}
			
			for (int i = 0; i<=lastrank; i++)
			{
				System.out.print(ranknumtest[i] + "  ");
			}
			
			System.out.println(getNumberOfEvaluations());
			//test code end
			
			
			
		}
		else
		{
			
			boolean offspringok = false;

			while (offspringok == false)
			{
				selection = new TournamentSelection(
						new NondominatedFitnessComparator2());

				Solution[] parents = selection.select(variation.getArity(), population);
				Solution[] offspring = variation.evolve(parents);
				
				evaluate(offspring[0]);
				offspring[0].setAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE, Double.NEGATIVE_INFINITY);
				population.add(offspring[0]);
				
				// rank the solutions and calculate crowding distance
				new FastNondominatedSorting().evaluate(population);

				population.sort(new NondominatedFitnessComparator());
				
				solution = population.get(populationSize);
				lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
				
				int lastrank2 = (int)population.get(populationSize-1).getAttribute(RANK_ATTRIBUTE);
					

				if (lastrank == 0)
				{
					fitnessEvaluator.evaluate(population);
					offspringok = true;
				}else if (lastrank2 != 0)
				{
					computeFitnessForLastFront();
					offspringok = true;
				}else
				{
					population.remove(populationSize);
				}
				
			}

			
			population.truncate(populationSize, 
					new NondominatedFitnessComparator2());
			
			
			solution = population.get(populationSize-1);
			lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
			
//			if ((lastrank != 0))
//			{
//				selection = new TournamentSelection(
//						new NondominatedFitnessComparator());
//			}
			
//			//debug code
//			for (int i = 0; i<populationSize; i++)
//			{
//				if (population.get(i).getAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE) == null)
//				{
//					System.out.print(i);
//				}
//				
//			}
			
			
			//test code
			solution = population.get(populationSize-1);
			lastrank = (int) solution.getAttribute(RANK_ATTRIBUTE);
			int[] ranknumtest = new int[lastrank+1];
			for (int i = 0; i<populationSize; i++)
			{
				int temp = (int) population.get(i).getAttribute(RANK_ATTRIBUTE);
				ranknumtest[temp]++; 
			}
			
			for (int i = 0; i<=lastrank; i++)
			{
				System.out.print(ranknumtest[i] + "  ");
			}
			
			System.out.println(getNumberOfEvaluations());
			//test code end
			
			
			
		}
		
		
		
		//TEST CODE
//		int[] ranknum = new int[lastrank+1];
//		for (int i = 0; i<populationSize; i++)
//		{
//			int temp = (int) population.get(i).getAttribute(RANK_ATTRIBUTE);
//			ranknum[temp]++; 
//		}
//	
//		for (int i = 0; i<=lastrank; i++)
//		{
//			System.out.print(ranknum[i] + "  ");
//		}
//		
//		System.out.println(getNumberOfEvaluations());
		//TEST CODE END

	}
	
	/**
	 * Computes the fitness for the last front.
	 */
	private void computeFitnessForLastFront() {
		Population front = new Population();
		int rank = 1;
		
		for (Solution solution : population) {
			int solutionRank = (Integer)solution.getAttribute(
					FastNondominatedSorting.RANK_ATTRIBUTE);
			
			if (solutionRank > rank) {
				front.clear();
				rank = solutionRank;
			}
			
			if (solutionRank == rank) {
				front.add(solution);
			}
			
			solution.setAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE, 0.0);
		}
		
		fitnessEvaluator.evaluate(front);
	}

	
	private class NondominatedFitnessComparator extends ChainedComparator
	implements Comparator<Solution>, Serializable {

		private static final long serialVersionUID = -4088873547790962685L;

		public NondominatedFitnessComparator() {
//			super(new RankComparator(), new FitnessComparator(
//					fitnessEvaluator.areLargerValuesPreferred()));
			super(new RankComparator());

		}

	}
	
	private class NondominatedFitnessComparator1 extends ChainedComparator
	implements Comparator<Solution>, Serializable {

		private static final long serialVersionUID = -4088873647790962685L;

		public NondominatedFitnessComparator1() {
			super(new RankComparator(), new CrowdingComparator());
		}

	}
	
	
	private class NondominatedFitnessComparator2 extends ChainedComparator
	implements Comparator<Solution>, Serializable {

		private static final long serialVersionUID = -4088873547760962685L;

		public NondominatedFitnessComparator2() {
			super(new RankComparator(), new FitnessComparator(
					fitnessEvaluator.areLargerValuesPreferred()));

		}

	}

}
