package org.moeaframework.algorithm;

import java.io.Serializable;
import java.util.Comparator;

import org.moeaframework.core.FastNondominatedSorting;
import org.moeaframework.core.FitnessEvaluator;
import org.moeaframework.core.Initialization;

import org.moeaframework.core.Population;
import org.moeaframework.core.Problem;
import org.moeaframework.core.Selection;
import org.moeaframework.core.Solution;
import org.moeaframework.core.Variation;
import org.moeaframework.core.comparator.ChainedComparator;

import org.moeaframework.core.comparator.FitnessComparator;
import org.moeaframework.core.comparator.NondominatedSortingComparator;

import org.moeaframework.core.comparator.RankComparator;
import org.moeaframework.core.operator.TournamentSelection;



public class HSMS3 extends AbstractEvolutionaryAlgorithm {

	
	/**
	 * The fitness evaluator to use (e.g., hypervolume or additive-epsilon
	 * indicator).
	 */
	private FitnessEvaluator fitnessEvaluator;
	
	/**
	 * The selection operator.
	 */
	private Selection selection;
	
	/**
	 * The variation operator.
	 */
	private Variation variation;

	//private boolean initial;
	
	/**
	 * Constructs a new SMS-EMOA instance.
	 * 
	 * @param problem the problem
	 * @param initialization the initialization operator
	 * @param variation the variation operator
	 * @param fitnessEvaluator the fitness evaluator
	 */
	public HSMS3(Problem problem, Initialization initialization,
			Variation variation, FitnessEvaluator fitnessEvaluator) {
		super(problem,
				new Population(),
				null,
				initialization);
		this.variation = variation;
		this.fitnessEvaluator = fitnessEvaluator;
		//this.initial = false;
		
		if (fitnessEvaluator ==  null) {
			selection = new TournamentSelection(
					new NondominatedSortingComparator());
		} else {
			selection = new TournamentSelection(
					new NondominatedFitnessComparator2());
		}

	}

	@Override
	protected void initialize() {
		super.initialize();
		
		// rank the solutions
		//rank and crowding distance(for removing duplicated solutions)
		new FastNondominatedSorting().evaluate(population);
		//initial = false;
		
		//calculate  contribution
//		if (fitnessEvaluator != null) {
			fitnessEvaluator.evaluate(population);
//		}
	}

	@Override
	protected void iterate() {

		int populationSize = population.size();
		Solution[] parents = selection.select(variation.getArity(), population);
		Solution[] offspring = variation.evolve(parents);
		
		evaluate(offspring[0]);
		offspring[0].setAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE, 0.0);
		population.add(offspring[0]);
		// rank the solutions and calculate crowding distance
		new FastNondominatedSorting().evaluate(population);
		
		computeFitnessForLastFront();
		
		population.truncate(populationSize, 
				new NondominatedFitnessComparator2());

		
		
		
		//TEST CODE
//		int[] ranknum = new int[lastrank+1];
//		for (int i = 0; i<populationSize; i++)
//		{
//			int temp = (int) population.get(i).getAttribute(RANK_ATTRIBUTE);
//			ranknum[temp]++; 
//		}
//	
//		for (int i = 0; i<=lastrank; i++)
//		{
//			System.out.print(ranknum[i] + "  ");
//		}
//		
//		System.out.println(getNumberOfEvaluations());
		//TEST CODE END

	}
	
	/**
	 * Computes the fitness for the last front.
	 */
	private void computeFitnessForLastFront() {
		Population front = new Population();
		int rank = 1;
		
		for (Solution solution : population) {
			int solutionRank = (Integer)solution.getAttribute(
					FastNondominatedSorting.RANK_ATTRIBUTE);
			
			if (solutionRank > rank) {
				front.clear();
				rank = solutionRank;
			}
			
			if (solutionRank == rank) {
				front.add(solution);
			}
			
			//solution.setAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE, 0.0);
		}
		
		fitnessEvaluator.evaluate(front);
	}

	
//	private class NondominatedFitnessComparator extends ChainedComparator
//	implements Comparator<Solution>, Serializable {
//
//		private static final long serialVersionUID = -4088873547790962685L;
//
//		public NondominatedFitnessComparator() {
////			super(new RankComparator(), new FitnessComparator(
////					fitnessEvaluator.areLargerValuesPreferred()));
//			super(new RankComparator());
//
//		}
//
//	}
	
//	private class NondominatedFitnessComparator1 extends ChainedComparator
//	implements Comparator<Solution>, Serializable {
//
//		private static final long serialVersionUID = -4088873647790962685L;
//
//		public NondominatedFitnessComparator1() {
//			super(new RankComparator(), new CrowdingComparator());
//		}
//
//	}
	
	
	private class NondominatedFitnessComparator2 extends ChainedComparator
	implements Comparator<Solution>, Serializable {

		private static final long serialVersionUID = -4088873547760962685L;

		public NondominatedFitnessComparator2() {
			super(new RankComparator(), new FitnessComparator(
					fitnessEvaluator.areLargerValuesPreferred()));

		}

	}

}
